#
# Apply patch $patchFile, located in directory $patchBaseDir to tree at $treeToPatch
#
# Patches located in subdirectories of $patchBaseDir are applied in the corresponding 
# subdirectory of $treeToPatch. As usual patches are generated by diff-ing two subtrees.
#
# Record the patches once applied in $patchStatusDir
#
: ${PATCHES_LOC:=$recipe_metadatadir/patches}

apply_patches() {
	local treeToPatch="$1"
	shift
	local patchBaseDir="$1"
	shift
	local patchStatusBase="$1"
	shift
	local patchStatusBaseDir="$patchStatusBase/applied_patches"
	local patchBackupBaseDir="$patchStatusBase/backups"
	local patchFile
	
	for patchFile; do
		local parentFolder=$(dirname "$patchFile")
		local patchName=$(basename "$patchFile")
		local targetDir="$treeToPatch/$parentFolder"
		local patchStatusDir="$patchStatusBaseDir/$parentFolder"
		local patchStatus="$patchStatusBaseDir/$patchFile"
		local patchBackupDir="$patchBackupBaseDir/$parentFolder"
		local patchBackup="$patchBackupBaseDir/$patchFile"
		
		# Apply patch
		mkdir -p "$treeToPatch/$parentFolder"
		
		# Make backups in a specific subtree to allow reliably reverting patches; 
		# problem is that new files are neither backed-up nor revertible, so we must
		# also keep a list fo touched files...
		rm -rf "$patchStatus" "$patchBackup"
		mkdir -p "$patchBackup"
		sed -n -e 's|---[[:space:]][^/]\+/\([^[:space:]]\+\)[[:space:]].*$|\1|p' \
		       -e 's|+++[:space:][^/]\+/\([^[:space:]]\+\)[[:space:]].*$|\1|p' < "$patchBaseDir/$patchFile" \
		| sort | uniq > "$patchBackup.changed_files"
		#echo "patch -N -b -V simple -B "$patchBackup/" -d '$treeToPatch/$parentFolder' -p1 < '$patchBaseDir/$patchFile'"
		if patch -N -b -V simple -B "$patchBackup/" -d "$treeToPatch/$parentFolder" -p1 < "$patchBaseDir/$patchFile"; then
			# Now record the applied patch 
			mkdir -p "$patchStatusDir"
			cp "$patchBaseDir/$patchFile" "$patchStatus"
		else
			# Here we should unapply the not-fully-applied patch:
			# First remove modified patches
			(cd "$targetDir"; rm -f $(cat "$patchBackup.changed_files"))
			# Then reinstall backupped files
			tar -cf - -C "$patchBackup" . | tar xf - -C "$targetDir"
			# And clean up patch status
			rm -rf "$patchBackup.changed_files" "$patchBackup"
			return 1
		fi
		echo "Applied patch '$patchFile'"
	done
}

unapply_patches() {
	local treeToPatch="$1"
	shift
	local patchStatusBase="$1"
	shift
	local patchStatusBaseDir="$patchStatusBase/applied_patches"
	local patchBackupBaseDir="$patchStatusBase/backups"
	local patchFile
	
	for patchFile; do
		local parentFolder=$(dirname "$patchFile")
		local patchName=$(basename "$patchFile")
		local targetDir="$treeToPatch/$parentFolder"
		local patchStatusDir="$patchStatusBaseDir/$parentFolder"
		local patchStatus="$patchStatusBaseDir/$patchFile"
		local patchBackupDir="$patchBackupBaseDir/$parentFolder"
		local patchBackup="$patchBackupBaseDir/$patchFile"

		if [ -r "$patchBackup.changed_files" -a -d "$patchBackup" ]; then
			# Revert the patch by using backups
			# First remove modified patches
			(cd "$targetDir"; rm -f $(cat "$patchBackup.changed_files"))
			# Then reinstall backupped files
			tar -cf - -C "$patchBackup" . | tar xf - -C "$targetDir"
		else
			# No available backup, apply patch "reversed"
			#echo "patch -E -R -d '$treeToPatch/$parentFolder' -p1 < '$patchStatus'"
			patch -E -R -d "$treeToPatch/$parentFolder" -p1 < "$patchStatus"
		fi
	
		# Now record the patch was no more present
		rm -rf "$patchStatus" "$patchBackup.changed_files" "$patchBackup"
		echo "Removed patch '$patchFile'"
	done
}

patch_recipe() {
	local treeToPatch="$1"
	local patchBaseDir="$2"
	case $treeToPatch in
		/*)		;;
		.)		treeToPatch="$here" ;;
		./*)	treeToPatch="$here/${treeToPatch#./}" ;;
		*)		treeToPatch="$here/$treeToPatch" ;;
	esac
	case $patchBaseDir in
		/*)		;;
		.)		patchBaseDir="$here" ;;
		./*)	patchBaseDir="$here/${patchBaseDir#./}" ;;
		*)		patchBaseDir="$here/$patchBaseDir" ;;
	esac
	local patchStatusBase="${3:-$treeToPatch/.ac6_build}"
	case $patchStatusBase in
		/*)		;;
		.)		patchStatusBase="$here" ;;
		./*)	patchStatusBase="$here/${patchStatusBase#./}" ;;
		*)		patchStatusBase="$here/$patchStatusBase" ;;
	esac
	local patchFile allPatches patchesToClean
	local patchStatusDir="$patchStatusBase/applied_patches"

	if [ ! -d "$patchBaseDir" ]; then return 0; fi

	for patchFile in $(cd "$patchBaseDir"; find . -name "*.patch" | sed -e 's|^./||' | sort); do 
		if [ -r "$patchBaseDir/$patchFile" ]; then 
			allPatches="$allPatches ${patchFile#./}"
		fi
	done
	mkdir -p "$patchStatusDir"
	for patchFile in $(cd "$patchStatusDir"; find . -name "*.patch" | sed -e 's|^./||' | sort); do 
		if [ -r "$patchStatusDir/$patchFile" ]; then 
			if cmp -s "$patchStatusDir/$patchFile" "$patchBaseDir/$patchFile"; then
				appliedPatches="$appliedPatches ${patchFile#./}"
			else
				# The patch to apply was different from the already applied one, so
				# don't consider it as already applied, but note something was done
				# on the source code, so all applied patches will have to be unapplied
				appliedPatches="$appliedPatches ${patchFile#./}.modified"
			fi
			patchesToClean="${patchFile#./} $patchesToClean"
		fi
	done
	addedPatches="${allPatches#$appliedPatches}"
	if [ "$addedPatches" = "$allPatches" ]; then
		#
		# The set of patches to be applied is not an exact superset of the set of 
		# already applied patches, so we must suppress all applied patches to 
		# restore the source tree in its original state, then apply all patches
		#
		# It should be possible to be a bit more clever and keeop the common set of 
		# inital patches, unapplying the following already applied patches and applying
		# the needed new patches but this should only occur very rarely so why bother?
		#
		if [ -n "$patchesToClean" ]; then
			echo "Remove already applied patches"
			unapply_patches "$treeToPatch" "$patchStatusBase" $patchesToClean
			echo
		fi
	else
		#
		# We are just adding new patches at the end of the set of existing patches
		# so we can just apply the new patches on the already patched source tree
		#
		allPatches="$addedPatches"
	fi
	if [ -n "$allPatches" ]; then
		echo "Apply patches"
		apply_patches "$treeToPatch" "$patchBaseDir" "$patchStatusBase" $allPatches
	fi
}
